﻿using CySmart.Common.Base.Compatibility.Net20;
using CySmart.DongleCommunicator.API;
using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace ApiExamples.Scanner
{
    #region Scanner

    /// <summary>
    /// Simple class which logs all discovered BLE devices to the Console
    /// </summary>
    class Scanner : IDisposable
    {
        #region Setup Constants

        // Modify the constant values to match with your setup
        const string COM_PORT_NAME = "COM56";

        #endregion

        #region members

        // CySmart dongle communicator instance
        ICySmartDongleCommunicator m_communicator;

        // BLE manager callback
        BleMgrCb m_bleMgrCb;

        // Scan callback
        ScanCb m_scanCb;

        #endregion

        #region props

        /// <summary>
        /// Gets the BLE manager
        /// </summary>
        ICyBleMgr BleMgr
        {
            get { return m_communicator.BleMgr; }
        }

        #endregion

        #region ctor

        public Scanner()
        {
            m_communicator = null;
            m_bleMgrCb = null;
        }

        #endregion

        #region public methods

        /// <summary>
        /// Connect to CySmart BLE dongle
        /// </summary>
        public CyApiErr ConnectToDongle(CyDongleInfo info)
        {
            CySmartDongleMgr dongleMgr = CySmartDongleMgr.GetInstance();
            CyApiErr err = dongleMgr.TryGetCySmartDongleCommunicator(info, out m_communicator);
            if (err.IsOK)
            {
                m_bleMgrCb = new BleMgrCb();
                err = BleMgr.RegisterBleMgrCallback(m_bleMgrCb);

                m_scanCb = new ScanCb();
                SetupScanResultHandler();
            }

            return err;
        }

        /// <summary>
        /// Start scan
        /// </summary>
        public CyApiErr StartScan()
        {
            return StartScanHelper();
        }

        /// <summary>
        /// Stop scan
        /// </summary>
        public CyApiErr StopScan()
        {
            return StopScanHelper();
        }

        #endregion

        #region private

        /// <summary>
        /// Setup the scan result handler to log the discovered BLE devices
        /// </summary>
        private void SetupScanResultHandler()
        {
            m_scanCb.ScanResultHandler = (result) =>
            {
                if (result != null)
                {
                    foreach (var item in result.ScanRecords)
                    {
                        Console.WriteLine("Peer device: [{0:X12}, {1}], ADV_TYPE: {2}, RSSI: {3}",
                            item.PeerDeviceAddress.Address,
                            item.PeerDeviceAddress.AddressType,
                            item.AdvertisementType,
                            item.RSSI);
                    }
                }
            };
        }

        /// <summary>
        /// Start scan helper
        /// </summary>
        private CyApiErr StartScanHelper()
        {
            AutoResetEvent sync = new AutoResetEvent(false);
            CyApiErr err = CyApiErr.OK;

            // Setup the scan status changed handler
            m_scanCb.ScanStatusChangedHandler = (status) =>
            {
                Console.WriteLine("Scan staus: {0}", status);
                if (status != CyScanStatus.IN_PROGRESS)
                    err = new CyApiErr("Failed to start the scan");

                sync.Set();
            };

            err = BleMgr.StartScan(new CyBleScanSettings(), m_scanCb);
            if (err.IsOK)
                sync.WaitOne();

            return err;
        }

        /// <summary>
        /// Stop scan helper
        /// </summary>
        private CyApiErr StopScanHelper()
        {
            AutoResetEvent sync = new AutoResetEvent(false);
            CyApiErr err = CyApiErr.OK;

            // Setup the scan status changed handler
            m_scanCb.ScanStatusChangedHandler = (status) =>
            {
                Console.WriteLine("Scan staus: {0}", status);
                if (status != CyScanStatus.STOPPED)
                    err = new CyApiErr("Failed to stop the scan");

                sync.Set();
            };

            err = BleMgr.StopScan();
            if (err.IsOK)
                sync.WaitOne();

            return err;
        }

        #endregion

        #region IDisposable

        /// <summary>
        /// Dispose the notification monitor instance
        /// </summary>
        public void Dispose()
        {
            if (m_communicator != null)
            {
                CyApiErr err = CySmartDongleMgr.GetInstance().CloseCommunicator(m_communicator);
                if (err.IsNotOK)
                    Console.WriteLine(err.Message);

                m_communicator = null;
            }
        }

        #endregion

        #region Main

        static void Main(string[] args)
        {
            using (Scanner scanner = new Scanner())
            {
                Console.WriteLine("Connecting to the dongle at {0} ...", COM_PORT_NAME);

                // Change the dongle type to match with the dongle you need to connect
                CyApiErr err = scanner.ConnectToDongle(new CyDongleInfo(COM_PORT_NAME, CyDongleInfo.CySmartDongleType.CY5670));
                if (err.IsNotOK)
                {
                    Console.WriteLine(err.Message);
                    return;
                }

                Console.WriteLine("Start scan ...");
                err = scanner.StartScan();
                if (err.IsNotOK)
                {
                    Console.WriteLine(err.Message);
                    return;
                }

                Console.WriteLine("Press any key to stop scan");
                Console.ReadKey();

                Console.WriteLine("Stop scan ...");
                err = scanner.StopScan();
                if (err.IsNotOK)
                {
                    Console.WriteLine(err.Message);
                    return;
                }
            }
        }

        #endregion
    }

    #endregion

    #region BleMgrCb

    /// <summary>
    /// BLE manager Callback class
    /// </summary>
    class BleMgrCb : CyBleMgrCallback
    {
        /// <summary>
        /// Gets/Sets the connection handler
        /// </summary>
        public Action<CyConnectResult, CyStatus> ConnectionHandler { get; set; }

        public override void OnConnected(CyConnectResult result, CyStatus status)
        {
            if (ConnectionHandler != null)
                ConnectionHandler(result, status);
        }
    }

    #endregion

    #region ScanCb

    /// <summary>
    /// Scan callback class
    /// </summary>
    class ScanCb : CyScanCallback
    {
        #region props

        /// <summary>
        /// Gets/Sets the scan result handler
        /// </summary>
        public Action<CyScanResult> ScanResultHandler { get; set; }

        /// <summary>
        /// Gets/Sets the scan status changed handler
        /// </summary>
        public Action<CyScanStatus> ScanStatusChangedHandler { get; set; }

        #endregion

        #region overrides

        public override void OnScanResult(CyScanResult result)
        {
            if (ScanResultHandler != null)
                ScanResultHandler(result);
        }

        public override void OnScanStatusChanged(CyScanStatus scanStatus)
        {
            if (ScanStatusChangedHandler != null)
                ScanStatusChangedHandler(scanStatus);
        }

        #endregion
    }

    #endregion
}
